#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `osc_recorder` package."""

import pytest

from osc_recorder import TimeSeriesBundle

pytest_plugins = 'pytester'

@pytest.fixture()
def ts_bundle():
    """Loads a time series bundle for test use."""
    return TimeSeriesBundle('tests/sine60.json')



"""
OSC input/output utility.
Server implementation to capture OSC.
Function implementation to replay OSC.
"""
import time
import logging
import numpy as np

import liblo

from .base import TimeSeriesBundle
from .collector import Collector

log = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)


# *********************************SERVER CLASS**********************************

class CaptureOSCServer(liblo.ServerThread):
    """OSC server to capture OSC data and store.

        Args:
            port (int): Port to listen on.

        Keyword Args:
            debug (bool): Set to True to log captures.
    """

    def __init__(self, port, debug=False):
        super().__init__(port)
        if debug:
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
        log.info("Server started on port %d", port)
        self.collector = Collector()
        self.first_message_captured = False

    @liblo.make_method(None, None)
    def catch_all_callback(self, path, args):
        """Executes when capturing any OSC message.

        Uses Collector to format and store.

        Args:
            path (string): OSC path captured on.
            args (list): list of values to capture.
        """
        log.debug("%s, %s", path, str(args))
        if not self.first_message_captured:
            self.first_message_captured = True
        self.collector.collect_samples(path, liblo.time(), args)

    def get_time_series(self):
        """Converts captured OSC messages to time series bundle format.

        Returns:
            (TimeSeriesBundle): All outputs are TimeSeriesBundles.
        """
        log.info("Converting Data...")
        return self.collector.to_time_series_bundle()


def replay_osc(data, address='127.0.0.1', port=9000, loop=False, debug=False):
    """Replay TimeSeriesBundle data via OSC by waiting.

    Calculates wait time between OSC messages and waits to send message.

    Args:
        data (TimeSeriesBundle): TimeSeriesBundle data.

    Keyword Args:
        address (str): IP address to send to.
        port (int): port to send to.
        loop (bool): set to True to repeat infinitely.
        debug (bool): set to True to print messages.

    """
    try:
        output_thread = liblo.Address(address, port)
        print(f'Replaying OSC to {address}:{port}')
    except liblo.AddressError as err:
        print(str(err))

    if not isinstance(data, TimeSeriesBundle):
        raise "Data type not supported."

    ts_dict = data.to_timestamped_dict()
    timestamps = sorted(ts_dict.keys())
    wait_times = np.diff(timestamps, append=0)

    run_once = True  # guarantee the selection runs once
    while run_once or loop:  # if loop selected, repeat the full replay
        run_once = False

        overhead_time_start = time.time()
        for i, timestamp in enumerate(timestamps):
            msg_list = list()
            for info in ts_dict[timestamp]:
                name = info[0]
                samples = info[1]
                if debug:
                    print("%s:%d -- %f %s %s" %
                          (address, port, timestamp, name, str(samples)))
                msg_list.append(liblo.Message(name, *samples))

            liblo.send(output_thread, *msg_list)

            overhead_time = time.time() - overhead_time_start
            # Avoids discontinous jumps in timestamps
            if wait_times[i] > overhead_time and wait_times[i] < 50:
                # NOTE:
                    # The wait time is still too slow
                    # Manual decrease in wait time
                    # check corresponding jupyter notebook
                time.sleep((wait_times[i] - overhead_time)* 0.8)
            # Starting overhead time immidiately after last call.
            overhead_time_start = time.time()

def replay_osc_time_bundle(data, address='127.0.0.1', port=9000, loop=False, debug=False):
    """Replay TimeSeriesBundle data via OSC. Doesn't wait.

    Sends all the packets at once with the appropriate wait times added to each time tag.
    Supported by many osc libraries that implemented the latest osc protocol update.
    The client can send packets with future timestamps. The server will wait till the
    correct time to actually call the callback functions.

    Args:
        data (TimeSeriesBundle): TimeSeriesBundle data.

    Keyword Args:
        address (str): IP address to send to.
        port (int): port to send to.
        loop (bool): set to True to repeat infinitely.
        debug (bool): set to True to print messages.

    """
    try:
        output_thread = liblo.Address(address, port)
        print(f'Replaying OSC to {address}:{port}')
    except liblo.AddressError as err:
        print(str(err))

    if not isinstance(data, TimeSeriesBundle):
        raise "Data type not supported."

    ts_dict = data.to_timestamped_dict()
    timestamps = sorted(ts_dict.keys())
    start_timestamp = timestamps[0]

    run_once = True  # guarantee the selection runs once
    while run_once or loop:  # if loop selected, repeat the full replay
        run_once = False

        bundle_list = list()
        start_time = liblo.time()
        for timestamp in timestamps:
            msg_list = list()
            for info in ts_dict[timestamp]:
                name = info[0]
                samples = info[1]
                if debug:
                    print("%s:%d -- %f %s %s" %
                          (address, port, timestamp, name, str(samples)))
                msg_list.append(liblo.Message(name, *samples))
            wait_time = timestamp - start_timestamp
            bundle = liblo.Bundle(start_time + wait_time, *msg_list)
            bundle_list.append(bundle)
        liblo.send(output_thread, *bundle_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Replay OSC signals from a file in current directory.
usage: python replay_osc.py -f test.json -a 127.0.0.1 -p 9000 -d
"""

import os
import argparse

from osc_recorder import TimeSeriesBundle, replay_osc, replay_osc_time_bundle

parser = argparse.ArgumentParser(
    description='Replay OSC data stream from file')
parser.add_argument('-a', '--address',
                    dest='address', type=str, default='127.0.0.1',
                    help='OSC address to replay to.')
parser.add_argument('-p', '--port',
                    dest='port', type=int, default=9000,
                    help='OSC port to listen on.')
parser.add_argument('-l', '--loop',
                    action='store_true', dest='loop',
                    help='Replay over and over.')
parser.add_argument('-d', '--debug',
                    action='store_true', dest='debug',
                    help='Print values being replayed.')
parser.add_argument('-t', '--time-bundle',
                    action='store_true', dest='time_bundle',
                    help='Sends using timestamped bundle format.')
requiredNamed = parser.add_argument_group('required arguments')
requiredNamed.add_argument('-f', '--filename',
                           dest='filename', type=str, required=True,
                           help='Name of file to replay.')
args = parser.parse_args()

file_path = os.path.abspath(args.filename)
data = TimeSeriesBundle()
data.from_json(file_path)


if args.time_bundle:
    replay_osc_time_bundle(data, address=args.address, port=args.port,
                           loop=args.loop, debug=args.debug)
else:
    replay_osc(data, address=args.address, port=args.port,
               loop=args.loop, debug=args.debug)

print('Done')



"""Base data class to store messaged from OSC paths."""
import json
from collections import defaultdict
import numpy as np


class TimeSeriesBundle():
    """Bundles together different path identified time series.
    The bundle is a dictionary with the path names as the key.
    The values in the bundle are also dictionaries.
    They contain the timestamps and samples which are stored in lists.
    Timestamps is a list of floats.
    Samples is a list of lists since there could be many channels.
    Example:
        bundle = {
            'path/name':{
                'timestamps': [ 0,   1,   2,  ...],
                'samples':    [[0], [1], [2], ...]
            }
        }

    Keyword Args:
        file_name (string): Import from a file if name provided.
    """
    def __init__(self, file_name=None):
        self.bundle = dict()
        if file_name:
            self.from_json(file_name)

    def add_time_series(self, name, timestamps, samples):
        """Adds a time series identified by a path name.

        Creates a dictionary of timestamps and values recieved for each path.

        Args:
            name (string): OSC path name.
            timestamps (float): Timestamp saved when packet is received.
            samples (list): List of values received.
        """
        if isinstance(timestamps, np.ndarray):
            timestamps = timestamps.tolist()
        if isinstance(samples, np.ndarray):
            samples = samples.tolist()
        self.bundle[name] = {
            'timestamps': timestamps,
            'samples': samples
        }

    def to_json(self, file_name):
        """Saves the bundle to a json file.

        Saved to the directory the script is called on.

        Args:
            file_name (string): File name to save as.
        """
        with open(file_name, 'w') as save_file:
            json.dump(self.bundle, save_file, indent=2)

    def from_json(self, file_name):
        """Replaces the current bundle with with the bundle from the json file.

        Args:
            file_name (string): File name to open.
        """
        with open(file_name, 'r') as read_file:
            self.bundle = json.load(read_file)

    def to_timestamped_dict(self):
        """Returns a dict with the timestamps as the key for the dictionary.

        timestamped_bundle = {
            timestamp0:[
                [ /name0,   /name1, ...],
                [    [0],      [1], ...]
            ],
            timestamp0:[
                [ /name0,   /name1, ...],
                [    [0],      [1], ...]
            ]

        }

        Returns:
            [type]: [description]
        """
        timestamped_dict = defaultdict(list)
        for name, info in self.bundle.items():
            for timestamp, sample in zip(info['timestamps'], info['samples']):
                timestamped_dict[timestamp].append([name, sample])
        return timestamped_dict



# -*- coding: utf-8 -*-

"""Top-level package for OSC-recorder."""


__version__ = '0.1.4'

from .base import TimeSeriesBundle
from .collector import Collector
from .osc_utils import CaptureOSCServer, replay_osc, replay_osc_time_bundle



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Capture OSC signals and save as a .json file in current directory.
Formated as a TimeSerieBundle.
usage: python replay_osc.py -p 9000 -f test -d
"""
import sys
import time
import argparse

import liblo

from osc_recorder import CaptureOSCServer

parser = argparse.ArgumentParser(
    description='Capture OSC and save to .json file in current directory.')
parser.add_argument('-p', '--port',
                    dest='port', type=int, default=9000,
                    help='OSC port to listen on.')
parser.add_argument('-f', '--filename',
                    dest='filename', type=str,
                    default='OSC_capture',
                    help='Name of json file to save to.')
parser.add_argument('-d', '--debug',
                    action='store_true', dest='debug',
                    help='Print values captured.')

args = parser.parse_args()

try:
    server = CaptureOSCServer(args.port, debug=args.debug)
    server.start()
    print("Server started on %d. Press Ctrl-C to terminate and save." % (args.port))
    print("Data saved to %s.json" % args.filename)
    if not args.debug:
        print("Debug is off. Can be turned on with '-d' flag.")
except liblo.ServerError as err:
    print(str(err))

while True:
    try:
        time.sleep(0.1)
    except KeyboardInterrupt:
        time_series_bundle = server.get_time_series()
        time_series_bundle.to_json(args.filename+'.json')
        server.free()
        print('Done')
        sys.exit()

sys.exit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `osc_recorder` package."""

# import pytest

from osc_recorder import TimeSeriesBundle

def test_time_series_bundle_format(ts_bundle):
    """Testing TimeSeriesBundle is in the correct format."""
    assert isinstance(ts_bundle, TimeSeriesBundle)
    for name, info in ts_bundle.bundle.items():
        assert isinstance(name, str)
        assert isinstance(info, dict)
        assert isinstance(info['timestamps'], list)
        assert isinstance(info['samples'], list)


def test_bundle_to_timestamped_dict(ts_bundle):
    """Testing TimeSeriesBundle conversion to timestampled_dict"""
    ts_dict = ts_bundle.to_timestamped_dict()

    for name, info in ts_bundle.bundle.items():
        timestamps = info['timestamps']
        samples = info['samples']
        for i, timestamp in enumerate(timestamps):
            sample = samples[i]
            assert timestamp in ts_dict
            assert [name, sample] in ts_dict[timestamp]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Convert an osc file to multiple csv files.

Accepts file.osc with contents:
____________________________________________________________________________________________________
osc_time          |path      |types  |packets                                                    
e17775e1.21044f19 |/muse/eeg |ffffff |825.201477 825.201477 825.201477 825.201477 nan 825.201477
e17775e1.21173fb7 |/muse/acc |fff    |0.000000 0.000000 0.000000
----------------------------------------------------------------------------------------------------

Parses and saves to multiple `path`.csv files.
The first column is the corresponding timestamp.

osc_time is in a weird format:
    x.y where x and y are hex numbers.
    x = seconds since 1 Jan 1900
    y = 2**-32 fraction of a second
    Subtract 2,208,988,800 seconds to get unix timestamp
"""
import os
from argparse import ArgumentParser
import numpy as np
from collections import defaultdict

eaxmple_usage = 'Usage: python osc_to_csv.py file.osc -f folder_path'

parser = ArgumentParser(
    description='Convert an osc file to multiple csv files.',
    epilog=eaxmple_usage)
parser.add_argument(
    'file_name', type=str,
    help='osc file captured via: ```oscdump [port] > file.osc```')
parser.add_argument(
    '-f', '--folder', type=str, dest='folder_path',
    help='A folder to place files in.')

args = parser.parse_args()
file_name = args.file_name
folder_path = args.folder_path

path_dict = defaultdict(lambda: [], {})

debug = False
num = 0

with open(file_name, 'r') as osc_file:
    for line in osc_file.readlines():
        line = line.strip('\n')
        osc_time, path, types, packets = line.split(' ', 3)
        x, y = osc_time.split('.')
        float_time = int(x, 16) + int(y,16)/2**32

        path_dict[path] += [[float_time] + [float(pack) for pack in packets.split(' ')]]

        # debug with smaller file size
        if debug:
            if num < 50:
                num = num + 1
            else:
                break

for path, path_list in path_dict.items():
    path = path[1::] if path[0] == '/' else path
    file_name = path.replace('/', '_')+'.csv'
    if folder_path:
        if not os.path.isdir(folder_path):
            os.makedirs(folder_path)
        file_name = os.path.join(folder_path, file_name)
    np.savetxt(file_name, np.array(path_list), delimiter=',')
    print(f'Saved: {file_name}')
print('Done!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
import sys

from distutils.command.build_scripts import build_scripts
from distutils import util, log

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['numpy', 'cython', 'pyliblo']

setup_requirements = []

test_requirements = ['pytest>=3.5.0']

class build_scripts_rename(build_scripts):
    def copy_scripts(self):
        build_scripts.copy_scripts(self)
        # remove the .py extension from scripts
        for s in self.scripts:
            f = util.convert_path(s)
            before = os.path.join(self.build_dir, os.path.basename(f))
            after = os.path.splitext(before)[0]
            log.info("renaming %s -> %s" % (before, after))
            os.rename(before, after)


cmdclass = {
    'build_scripts': build_scripts_rename,
}


setup(
    author="Oishe Farhan",
    author_email='oishe.farhan@interaxon.ca',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    description="Python scripts to capture and replay OSC packets. Built on top of pyliblo.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='osc_recorder',
    name='osc_recorder',
    packages=find_packages(include=['osc_recorder']),
    scripts=['scripts/capture_osc.py', 'scripts/replay_osc.py', 'scripts/osc_to_csv.py'],
    cmdclass=cmdclass,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/oishefarhan/osc_recorder',
    version='0.1.4',
    zip_safe=False,
)



"""A collector with simple storage for efficient callbacks during OSC packet capture."""
import copy
from collections import defaultdict

from .base import TimeSeriesBundle


class Collector():
    """Collector with a callback function to receive and store OSC packets."""
    def __init__(self):
        self._samples = defaultdict(list)
        self._timestamps = defaultdict(list)

    def collect_samples(self, name, timestamp, sample):
        """Callback function to collect samples from incoming OSC.

        Args:
            name (str): OSC path name.
            timestamp (float): timing information for each packet.
            sample (list, np.ndarray or scalar): values to collect.
        """
        self._timestamps[name].append(timestamp)
        self._samples[name].append(copy.deepcopy(sample))

    def to_time_series_bundle(self):
        """Converts OSC packets stored into the TimeSeriesBundle format.

        Returns:
            (TimeSeriesBundle): TimeSeriesBundleObject.
        """
        bundle = TimeSeriesBundle()
        for name in self._timestamps:
            bundle.add_time_series(name, self._timestamps[name], self._samples[name])
        return bundle
